/*  Copyright (C) 2005 Magnus Bjernstad <bjernstad@gmail.com>
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
using System;
using System.IO;
using System.Collections;
using Gtk;

namespace MpcSharp
{
    class Cover : ScrolledWindow, IUpdatable
    {
        enum ImageFormats
        {
            png,
            tiff,
            tif,
            jpg,
            jpeg,
            gif,
            pcx,
            ico,
            xpm,
            xbm
        }

        enum ZoomMode
        {
            NormalSize,
            BestFit
        }

        Connection conn;

        [Glade.Widget] MenuItem front;
        [Glade.Widget] MenuItem back;
        [Glade.Widget] MenuItem other;
        [Glade.Widget] MenuItem normal_size;
        [Glade.Widget] MenuItem best_fit;
        [Glade.Widget] Menu     cover_popup;
        Image image;



        // The dir in which the covers are
        string dir = "";
        // These are the filenames of the found covers
        string frontFile;
        string backFile;
        ArrayList otherList = new ArrayList();
        Gdk.Pixbuf pixbuf;

        // For panning of the cover image
        double oldXRoot;
        double oldYRoot;
        double oldHadj;
        double oldVadj;

        Connection.UpdateDelegate d;

        ZoomMode mode = ZoomMode.BestFit;
        ZoomMode Mode
        {
            get
            {
                return mode;
            }
            set
            {
                mode = value;
                QueueResize();
            }
        }

        string activeFile = "";
        string ActiveFile
        {
            get
            {
                return activeFile;
            }
            set
            {
                activeFile = value;
                LoadFile();
            }
        }



        public Cover(ref Connection conn) : base()
        {
            this.conn = conn;

            image = new Image();
            EventBox eventbox = new EventBox();
            eventbox.Visible = false;
            eventbox.Add(image);
            AddWithViewport(eventbox);
            SizeAllocated += new SizeAllocatedHandler(OnSizeAllocate);
            Mapped += new EventHandler(OnMapEvent);

            Glade.XML gxml_menu = Glade.XML.FromAssembly("cover.glade", "cover_popup", null);
            gxml_menu.Autoconnect(this);

            // Register update delegate
            d = new Connection.UpdateDelegate(Update);
            conn.RegisterUpdateDelegate(d);

            // GConf notifications
            MpcSharp.Settings.MusicDirectoryChanged += MusicDirectoryNotifyEventHandler;
        }

        public void UnregisterUpdateDelegate()
        {
            conn.UnregisterUpdateDelegate(d);
        }

        void Update()
        {
            if (conn.prevStatusInfo.Music.SongID !=
                    conn.statusInfo.Music.SongID)
            {
                // New song -> Find the images
                dir = System.IO.Path.GetDirectoryName(conn.musicInfo.Filename);
                FindImages();
            }
        }

        void MusicDirectoryNotifyEventHandler(object o, GConf.NotifyEventArgs args)
        {
            FindImages();
            QueueResize();
        }

        void FindImages()
        {
            // Zero previously found images
            ActiveFile = "";
            frontFile = "";
            backFile = "";
            otherList.Clear();

            if (MpcSharp.Settings.MusicDirectory == "" || dir == null)
                return;

            string absDir = System.IO.Path.Combine(MpcSharp.Settings.MusicDirectory, dir);
            string[] files;
            try
            {
                files = Directory.GetFiles(absDir);
            }
            catch (System.IO.DirectoryNotFoundException)
            {
                Console.WriteLine("Cover: Music directory not found");
                return;
            }
            ArrayList images = new ArrayList();
            foreach (string file in files)
            {
                try
                {
                    ImageFormats.Parse(typeof(ImageFormats),
                            System.IO.Path.GetExtension(file).TrimStart('.'), true);
                    images.Add(file);
                }
                catch (ArgumentException)
                {
                    // Not a match - do nothing
                }
            }

            // Extract front and back images.
            IEnumerator e = images.GetEnumerator();
            string imageFile;
            while (e.MoveNext())
            {
                imageFile = e.Current as string;
                if (imageFile.ToLower().IndexOf("front") >= 0)
                {
                    if (frontFile == "")
                    {
                        frontFile = imageFile;
                        continue;
                    }
                }
                else if (imageFile.ToLower().IndexOf("back") >= 0)
                {
                    if (backFile == "")
                    {
                        backFile = imageFile;
                        continue;
                    }
                }
                otherList.Add(imageFile);
            }
            otherList.Sort();

            // Show 'best' image
            if (frontFile != "")
                ActiveFile = frontFile;
            else if (backFile != "")
                ActiveFile = backFile;
            else if (otherList.Count > 0)
                ActiveFile = otherList[0] as string;

        }

        void LoadFile()
        {
            if (ActiveFile == "") // No cover file
                return;
            try
            {
                pixbuf = new Gdk.Pixbuf(ActiveFile);
                image.Pixbuf = pixbuf;
            }
            catch (GLib.GException)
            {
                Console.WriteLine("Cover: File not found: '{0}'", ActiveFile);
            }
            QueueResize();
        }

        /*
         * Main gui callbacks
         */
        //protected void OnSizeAllocate(Gdk.Rectangle allocation)
        protected void OnSizeAllocate(object o, System.EventArgs args)
        {
            // Abort if the covertab is not visible
            if ((Parent as Notebook).CurrentPageWidget != this)
            {
                return;
            }

            if (ActiveFile == "")
            {
                image.Pixbuf = new Gdk.Pixbuf(Gdk.Colorspace.Rgb, true, 8, 1, 1);
                image.Pixbuf.Fill(0);
                return;
            }

            if (Mode == ZoomMode.BestFit)
            {
                double width  = Hadjustment.PageSize;
                double height = Vadjustment. PageSize;
                double viewAR = width / height;
                double pixbufAR = pixbuf.Width / (double) pixbuf.Height;
                int newWidth, newHeight;
                if (viewAR < pixbufAR)
                {
                    newWidth = Convert.ToInt32(width);
                    newHeight = Convert.ToInt32(width / pixbufAR);
                }
                else
                {
                    newWidth = Convert.ToInt32(height * pixbufAR);
                    newHeight = Convert.ToInt32(height);
                }
                image.Pixbuf = 
                    pixbuf.ScaleSimple(newWidth, newHeight, Gdk.InterpType.Bilinear);
            }
            else
            { // Mode == ZoomMode.NormalSize
                image.Pixbuf = pixbuf;
            }
        }

        [GLib.ConnectBefore]
        protected override bool OnButtonPressEvent(Gdk.EventButton evnt)
        {
            if (evnt.Button == 3)
            {
                front.Sensitive = frontFile != "";
                back.Sensitive = backFile != "";
                other.Sensitive = otherList.Count > 0;
                normal_size.Sensitive = ActiveFile != "";
                best_fit.Sensitive = ActiveFile != "";

                // Construct menu for 'other' covers
                if (otherList.Count > 0)
                {
                    Menu menu = new Menu();
                    other.Submenu = menu;
                    IEnumerator e = otherList.GetEnumerator();
                    while (e.MoveNext())
                    {
                        string file = e.Current as string;
                        string label = System.IO.Path.GetFileName(file);
                        MenuItem item = new PayloadMenuItem(label, file);
                        menu.Append(item);
                        item.Activated += new EventHandler(OnMenuOther);
                    }
                    menu.ShowAll();
                }
                cover_popup.Popup();
            }
            else if (evnt.Button == 1)
            {
                oldXRoot = evnt.XRoot;
                oldYRoot = evnt.YRoot;
                oldHadj  = Hadjustment.Value;
                oldVadj  = Vadjustment.Value;
            }
            return false;
        }

        // Pan the displayed cover image with mouse button 1
        protected override bool OnMotionNotifyEvent(Gdk.EventMotion evnt)
        {
            // We must drag with mouse button 1...
            if ((evnt.State & Gdk.ModifierType.Button1Mask) !=
                    Gdk.ModifierType.Button1Mask)
                return false;
            // ...and be in 'NormalSize' mode
            if (Mode != ZoomMode.NormalSize)
                return false;

            // Calculate new Adjustment values based on where we began the dragging
            // motion.
            Adjustment hadj = Hadjustment;
            Adjustment vadj = Vadjustment;
            double dx = evnt.XRoot - oldXRoot;
            double dy = evnt.YRoot - oldYRoot;
            double newHadj = System.Math.Min(oldHadj - dx, hadj.Upper - hadj.PageSize);
            double newVadj = System.Math.Min(oldVadj - dy, vadj.Upper - vadj.PageSize);
            hadj.Value = newHadj;
            vadj.Value = newVadj;
            return false;
        }

        void OnMapEvent(object o, EventArgs args)
        {
            QueueResize();
        }

        /*
         * Menu callbacks
         */
        private void OnMenuFront(object o, EventArgs args)
        {
            ActiveFile = frontFile;
        }

        private void OnMenuBack(object o, EventArgs args)
        {
            ActiveFile = backFile;
        }

        private void OnMenuOther(object o, EventArgs args)
        {
            ActiveFile = (o as PayloadMenuItem).Payload as string;
        }

        private void OnMenuNormalSize(object o, EventArgs args)
        {
            Mode = ZoomMode.NormalSize;
        }

        private void OnMenuBestFit(object o, EventArgs args)
        {
            Mode = ZoomMode.BestFit;
        }

        /*
         * Extension of Gtk.MenuItem that can store an 'object' payload.
         */
        class PayloadMenuItem : Gtk.MenuItem
        {
            public object payload;
            public object Payload
            {
                get
                {
                    return payload;
                }
            }
            public PayloadMenuItem(string label, object payload) : base(label)
            {
                this.payload = payload;
            }
        }
    }
}
