/*  Copyright (C) 2005 Magnus Bjernstad <bjernstad@gmail.com>
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

using System;
using Gtk;
using MPD;

namespace MpcSharp
{
    public enum TimeFormat
    {
        Elapsed,
        Remaining
    }

    abstract class MpcSharpButton : Button
    {
        protected Connection conn;
        private Gdk.Pixbuf pixbuf;
        private Image image;
        protected string Resource
        {
            set
            {
                Gdk.Pixbuf pixbuf = new Gdk.Pixbuf(System.Reflection.Assembly.GetCallingAssembly(), value);
                image.Pixbuf = pixbuf;
                image.ShowAll();
            }
        }

        public MpcSharpButton(ref Connection conn) : base()
        {
            this.conn = conn;
            Relief = ReliefStyle.None;
            image = new Image();
            Add(image);
        }
    }

    abstract class MpcSharpToggleButton : ToggleButton
    {
        protected Connection conn;

        protected Gdk.Pixbuf pixbufActive;
        protected Gdk.Pixbuf pixbufInactive;
        protected Image image;

        protected string ActiveResource
        {
            set
            {
                pixbufActive = new Gdk.Pixbuf(System.Reflection.Assembly.GetCallingAssembly(), value);
            }
        }

        protected string InactiveResource
        {
            set
            {
                pixbufInactive = new Gdk.Pixbuf(System.Reflection.Assembly.GetCallingAssembly(), value);
            }
        }

        public MpcSharpToggleButton(ref Connection conn) : base()
        {
            this.conn = conn;
            Relief = ReliefStyle.None;
            image = new Image();
            Add(image);
        }
    }

    class PrevButton : MpcSharpButton
    {
        public PrevButton(ref Connection conn) : base(ref conn)
        {
            Resource = "prev.png";
        }

        override protected void OnClicked()
        {
                try
                {
                    conn.client.Prev();
                    conn.client.ReadResponse();
                }
                catch (MPD.MPDException e)
                {
                    Console.WriteLine(e);
                }
                catch (NullReferenceException)
                {}
        }
    }

    abstract class GuiWidget
    {
        protected Connection conn;

        public GuiWidget(ref Connection conn)
        {
            this.conn = conn;
        }

        // Return the widget that should be visible in the main gui
        public abstract Widget Contents
        {
            get;
        }
    }

    class PlayButton : MpcSharpButton
    {
        public PlayButton(ref Connection conn) : base(ref conn)
        {
            Resource = "play.png";
        }

        override protected void OnClicked()
        {
            StateStatus state = conn.statusInfo.State;
            try
            {
                if (state == StateStatus.Stop)
                    conn.client.Play();
                else
                    conn.client.Paused(state == StateStatus.Play ? true : false);
                conn.client.ReadResponse();
            }
            catch (MPD.MPDException e)
            {
                Console.WriteLine(e);
            }
            catch (NullReferenceException)
            {}
        }
    }

    class StopButton : MpcSharpButton
    {
        public StopButton(ref Connection conn) : base(ref conn)
        {
            Resource = "stop.png";
        }

        override protected void OnClicked()
        {
            try
            {
                conn.client.Stop();
                conn.client.ReadResponse();
            }
            catch (MPD.MPDException e)
            {
                Console.WriteLine(e);
            }
            catch (NullReferenceException)
            {}
        }
    }

    class NextButton : MpcSharpButton
    {
        public NextButton(ref Connection conn) : base(ref conn)
        {
            Resource = "next.png";
        }

        override protected void OnClicked()
        {
            try
            {
                conn.client.Next();
                conn.client.ReadResponse();
            }
            catch (MPD.MPDException e)
            {
                Console.WriteLine(e);
            }
            catch (NullReferenceException)
            {}
        }
    }

    class VolumeScale : HScale, IUpdatable
    {
        Connection.UpdateDelegate d;
        Connection conn;

        public VolumeScale(ref Connection conn) : base(0, 100, 2)
        {
            this.conn = conn;
            ValuePos = PositionType.Left;

            // Register update delegate
            d = new Connection.UpdateDelegate(Update);
            conn.RegisterUpdateDelegate(d);
        }
        public void UnregisterUpdateDelegate()
        {
            conn.UnregisterUpdateDelegate(d);
        }

        void Update()
        {
            if (Value != conn.statusInfo.Volume)
            {
                Value = conn.statusInfo.Volume;
            }
        }

        override protected void OnValueChanged()
        {
            try
            {
                conn.client.Volume((int) Value);
                conn.client.ReadResponse();
            }
            catch (MPD.MPDException e)
            {
                Console.WriteLine(e);
            }
            catch (NullReferenceException)
            {}
        }

        override protected string OnFormatValue(double val)
        {
            return "Volume:";
        }


    }

    class RandomButton : MpcSharpToggleButton, IUpdatable
    {
        Connection.UpdateDelegate d;

        public RandomButton(ref Connection conn) : base(ref conn)
        {
            ActiveResource = "random_on.png";
            InactiveResource = "random_off.png";
            image.Pixbuf = pixbufActive;
            Toggled += new EventHandler(OnToggled);

            // Register update delegate
            d = new Connection.UpdateDelegate(Update);
            conn.RegisterUpdateDelegate(d);
        }

        public void UnregisterUpdateDelegate()
        {
            conn.UnregisterUpdateDelegate(d);
        }

        void Update()
        {
            Toggled -= new EventHandler(OnToggled);
            Active = conn.statusInfo.Random;
            Toggled += new EventHandler(OnToggled);
            if (Active)
                image.Pixbuf = pixbufActive;
            else
                image.Pixbuf = pixbufInactive;
        }

        void OnToggled(object o, EventArgs args)
        {
            try
            {
                conn.client.Random(!conn.statusInfo.Random);
                conn.client.ReadResponse();
            }
            catch (MPD.MPDException e)
            {
                Console.WriteLine(e);
            }
            catch (NullReferenceException)
            {}
        }
    }

    class RepeatButton : MpcSharpToggleButton, IUpdatable
    {
        Connection.UpdateDelegate d;

        public RepeatButton(ref Connection conn) : base(ref conn)
        {
            ActiveResource = "repeat.png";
            Toggled += new EventHandler(OnToggled);
            image.Pixbuf = pixbufActive;

            // Register update delegate
            d = new Connection.UpdateDelegate(Update);
            conn.RegisterUpdateDelegate(d);
        }

        public void UnregisterUpdateDelegate()
        {
            conn.UnregisterUpdateDelegate(d);
        }

        void Update()
        {
            Toggled -= new EventHandler(OnToggled);
            Active = conn.statusInfo.Repeat;
            Toggled += new EventHandler(OnToggled);
        }

        void OnToggled(object o, EventArgs args)
        {
            try
            {
                conn.client.Repeat(!conn.statusInfo.Repeat);
                conn.client.ReadResponse();
            }
            catch (MPD.MPDException e)
            {
                Console.WriteLine(e);
            }
            catch (NullReferenceException)
            {}
        }
    }

    class TimeLabel : EventBox, IUpdatable
    {
        Connection conn;
        Connection.UpdateDelegate d;
        Label label;

        public TimeLabel(ref Connection conn) : base()
        {
            this.conn = conn;
            label = new Label();
            Add(label);

            // Register update delegate
            d = new Connection.UpdateDelegate(Update);
            conn.RegisterUpdateDelegate(d);
        }

        public void UnregisterUpdateDelegate()
        {
            conn.UnregisterUpdateDelegate(d);
        }

        void Update()
        {
            int elapsed = conn.statusInfo.Music.ElapsedTime;
            int total   = conn.statusInfo.Music.TotalTime;
            if (elapsed < 0)
                label.Text = "00:00 / 00:00";
            else if (MpcSharp.Settings.TimeFormat == TimeFormat.Elapsed)
            {
                label.Text = Format.TimeInSeconds(elapsed) + " / " +
                    Format.TimeInSeconds(total);
            }
            else if (MpcSharp.Settings.TimeFormat == TimeFormat.Remaining)
            {
                int remaining = System.Math.Abs(total - elapsed);
                label.Text = "-" + Format.TimeInSeconds(remaining) + " / " +
                    Format.TimeInSeconds(total);
            }
        }

        override protected bool OnButtonPressEvent(Gdk.EventButton evnt)
        {
            if (MpcSharp.Settings.TimeFormat == TimeFormat.Elapsed)
                MpcSharp.Settings.TimeFormat = TimeFormat.Remaining;
            else
                MpcSharp.Settings.TimeFormat = TimeFormat.Elapsed;
            return true;
        }
    }

    class ProgressScale : HScale, IUpdatable
    {
        Connection conn;
        Connection.UpdateDelegate d;

        public ProgressScale(ref Connection conn) : base(0, 1, 2)
        {
            this.conn = conn;
            //AdjustBounds += new AdjustBoundsHandler(OnAdjustBounds);
            DrawValue = false;

            // Register update delegate
            d = new Connection.UpdateDelegate(Update);
            conn.RegisterUpdateDelegate(d);
        }

        public void UnregisterUpdateDelegate()
        {
            conn.UnregisterUpdateDelegate(d);
        }

        void Update()
        {
            int elapsed = conn.statusInfo.Music.ElapsedTime;
            int total   = conn.statusInfo.Music.TotalTime;
            if (total > 0)
            {
                SetRange(0, total);
                Value = elapsed;
            }
            else
                Value = 0;
        }

        //void OnAdjustBounds(object o, AdjustBoundsArgs args)
        override protected void OnAdjustBounds(double new_value)
        {
            UpdatePolicy = Gtk.UpdateType.Discontinuous;
            try
            {
                conn.client.SeekTo(conn.statusInfo.Music.SongID,
                        Convert.ToInt32(new_value));
                conn.client.ReadResponse();
            }
            catch (MPD.MPDException e)
            {
                Console.WriteLine(e);
            }
            catch (NullReferenceException)
            {}
        }
    }
}
