/*  Copyright (C) 2005 Magnus Bjernstad <bjernstad@gmail.com>
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

using System;
using Gtk;
using MPD;


namespace MpcSharp
{
    public enum EntityType
    {
        File,
            Dir,
            Parent,
            Playlist
    }

    class Library : VBox, IUpdatable
    {
        Connection conn;
        System.Collections.Stack PositionStack = new System.Collections.Stack();
        Glade.XML gxml_menu;
        ListStore tree;
        string cwd = ""; // CurrentWorkingDir

        Label cwd_label = new Label();
        ScrolledWindow scrolled = new ScrolledWindow();
        TreeView view = new TreeView();
        Connection.UpdateDelegate d;


        void cellDataFunc (TreeViewColumn tvc, CellRenderer cell, TreeModel model, TreeIter iter)
        {
            int type = (int) model.GetValue(iter, 1);
            if (type == (int) EntityType.File)
                (cell as CellRendererPixbuf).StockId = Gtk.Stock.New;
            else if (type == (int) EntityType.Dir)
                (cell as CellRendererPixbuf).StockId = Gtk.Stock.Open;
            else if (type == (int) EntityType.Parent)
                (cell as CellRendererPixbuf).StockId = "";
            else if (type == (int) EntityType.Playlist)
                (cell as CellRendererPixbuf).StockId = Gtk.Stock.JustifyFill;
        }

        public Library(ref Connection conn) : base()
        {
            this.conn = conn;
            PackStart(cwd_label, false, false, 3);
            PackStart(scrolled);
            scrolled.Add(view);

            cwd_label.Xalign = 0;
            cwd_label.Xpad = 4;
            view.HeadersVisible = false;

            view.RowActivated += new RowActivatedHandler(OnRowActivated);
            view.ButtonPressEvent += new ButtonPressEventHandler(OnButtonPress);
            view.ButtonReleaseEvent += new ButtonReleaseEventHandler(OnButtonRelease);
            view.PopupMenu += new PopupMenuHandler(OnPopupMenu);

            gxml_menu = Glade.XML.FromAssembly("library.glade", "lib_popup", null);
            gxml_menu.Autoconnect(this);

            tree = new ListStore(typeof(string), typeof(int));

            view.Model = tree;
            CellRenderer cellText = new CellRendererText();
            CellRenderer cellPixbuf = new CellRendererPixbuf();

            view.AppendColumn("Type", cellPixbuf);
            view.Columns[0].SetCellDataFunc(cellPixbuf, 
                    new TreeCellDataFunc(cellDataFunc));
            view.AppendColumn("Library", cellText, "text", 0);
            view.Columns[1].Sizing = TreeViewColumnSizing.Autosize;
            view.Selection.Mode = SelectionMode.Multiple;

            // Register update delegate
            d = new Connection.UpdateDelegate(Update);
            conn.RegisterUpdateDelegate(d);
        }

        public void UnregisterUpdateDelegate()
        {
            conn.UnregisterUpdateDelegate(d);
        }

        // This function is called once for every poll to mpd
        void Update()
        {
            if (conn.statsInfo.DBUpdateDate > conn.prevStatsInfo.DBUpdateDate)
                Reload();
        }

        void Reload()
        {
            tree.Clear();
            cwd_label.Text = cwd;

            // Append parent dir entry if not at top already
            if (cwd != "")
            {
                tree.AppendValues("..", (int) EntityType.Parent);
            }

            try
            {
                CommandCallback callback;
                // Reload directories
                callback = new CommandCallback(ListDirectoriesCallback);
                conn.client.ListDirectories(cwd, callback);
                while (conn.client.ReadResponse())
                {}

                // Reload files
                callback = new CommandCallback(ListFilesCallback);
                conn.client.ListFiles(cwd, callback);
                while (conn.client.ReadResponse())
                {}
            }
            catch (MPD.MPDException e)
            {
                conn.client.CleanCache();
                Console.WriteLine(e);
            }
            catch (NullReferenceException)
            {}
        }

        void ListDirectoriesCallback(object result)
        {
            foreach (string dir in (string[]) result)
            {
                string[] temp = dir.Split('/');
                tree.AppendValues(temp[temp.Length - 1], (int) EntityType.Dir);
            }
        }

        void ListFilesCallback(object result)
        {
            foreach (string file in (string[]) result)
            {
                tree.AppendValues(System.IO.Path.GetFileName(file), (int) EntityType.File);

            }
        }

        void ChangeDir(TreePath treePath )
        {
            TreeIter iter = new TreeIter();
            tree.GetIter(out iter, treePath);

            string selected = (string) tree.GetValue(iter, 0);

            int type = (int) tree.GetValue(iter, 1);
            char delim = System.IO.Path.DirectorySeparatorChar;

            // Dir activated
            if (type == (int) EntityType.Dir)
            {
                PositionStack.Push(treePath.ToString());
                cwd = System.IO.Path.Combine(cwd, selected);
                Reload();
                view.ScrollToCell(new TreePath("0"), null, true, 0.5f, 0.0f);
            }
            // Parent dir activated
            else if (type == (int) EntityType.Parent)
            {
                try
                {
                    cwd = cwd.Substring(0, cwd.LastIndexOf(delim));
                }
                catch (System.ArgumentOutOfRangeException)
                {
                    cwd = "";
                }
                Reload();
                TreePath prevPath = new TreePath(PositionStack.Pop().ToString());
                view.SetCursor(prevPath, view.Columns[0], false);
                view.ScrollToCell(prevPath, null, true, 0.5f, 0.0f);
            }
            // File activated
            else if (type == (int) EntityType.File)
            {
                try
                {
                    conn.client.Add(System.IO.Path.Combine(cwd, selected));
                    conn.client.ReadResponse();
                }
                catch (MPD.MPDException)
                {}
                catch (NullReferenceException)
                {}
            }
        }

        void EnqueueSelection()
        {
            TreeIter iter;
            // Return if nothing is selected
            if (!tree.GetIterFirst(out iter))
                return;

            TreeModel model;
            TreePath[] paths = view.Selection.GetSelectedRows(out model);

            foreach (TreePath path in paths)
            {
                model.GetIter(out iter, path);
                EntityType type = (EntityType) model.GetValue(iter, 1);
                try
                {
                    if (type == EntityType.File || type == EntityType.Dir)
                    {
                        string val = (string) model.GetValue(iter, 0);
                        conn.client.Add(System.IO.Path.Combine(cwd, val)); }
                }
                catch (MPD.MPDException e)
                {
                    Console.WriteLine(e);
                }
                catch (NullReferenceException)
                {}
            }
            while (conn.client.ReadResponse()) {}
        }

        void ShowMenu()
        {
            bool selected = view.Selection.CountSelectedRows() > 0;
            gxml_menu.GetWidget("play").Sensitive = selected;
            gxml_menu.GetWidget("add").Sensitive = selected;

            (gxml_menu.GetWidget("lib_popup") as Menu).Popup();
        }

        /*
         * Main gui callbacks
         */
        void OnRowActivated(object o, RowActivatedArgs args)
        {
            ChangeDir(args.Path);
        }

        [GLib.ConnectBefore]
        void OnButtonPress(object o, ButtonPressEventArgs args)
        {
            if (args.Event.Button == 3)
            {
                // If multiple items selected, don't deselect them upon menu popup
                args.RetVal =  ((TreeView) o).Selection.CountSelectedRows() > 1;
            }
        }

        void OnButtonRelease(object o, ButtonReleaseEventArgs args)
        {
            if (args.Event.Button == 3)
            {
                ShowMenu();
            }
            else if (args.Event.Button == 2)
            {
                // Goto parent dir if not already in the root node
                TreeIter iter;
                tree.GetIter(out iter, new TreePath("0"));
                if ((EntityType) tree.GetValue(iter, 1) == EntityType.Parent)
                    ChangeDir(new TreePath("0"));
            }
            args.RetVal = true;
        }

        void OnPopupMenu(object o, PopupMenuArgs args)
        {
            ShowMenu();
        }

        /*
         * Menu callbacks
         */
        void OnMenuPlay(object o, EventArgs args)
        {
            try
            {
                conn.client.Clear();
                EnqueueSelection();
                conn.client.Play();
                conn.client.ReadResponse();
            }
            catch (MPD.MPDException e)
            {
                Console.WriteLine(e);
            }
            catch (NullReferenceException)
            {}
        }

        void OnMenuAdd(object o, EventArgs args)
        {
            EnqueueSelection();
            try
            {
                conn.client.Play();
                conn.client.ReadResponse();
            }
            catch (MPD.MPDException e)
            {
                Console.WriteLine(e);
            }
            catch (NullReferenceException)
            {}
        }

        void OnMenuUpdateAll(object o, EventArgs args)
        {
            try
            {
                conn.client.Update();
                conn.client.ReadResponse();
            }
            catch (MPD.MPDException e)
            {
                Console.WriteLine(e);
            }
            catch (NullReferenceException)
            {}
        }

        void OnMenuUpdateThisDir(object o, EventArgs args)
        {
            try
            {
                conn.client.Update(cwd);
                conn.client.ReadResponse();
            }
            catch (MPD.MPDException e)
            {
                Console.WriteLine(e);
            }
            catch (NullReferenceException)
            {}
        }

    }
}
