/*  Copyright (C) 2005 Magnus Bjernstad <bjernstad@gmail.com>
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

namespace MpcSharp
{

    using System;
    using System.Collections;
    using Gtk;
    using MPD;

    enum TabType
    {
        Playlist,
        Library,
        Find,
        Cover,
        Playlists
    }

    public enum MpcSharpWidget
    {
        Prev,
        Play,
        Stop,
        Next,
        Volume,
        Random,
        Repeat,
        Time,
        Progress
    }

    class MpcSharpApp
    {
        Connection conn;
        Glade.XML gxml;
        ArrayList visibleTabs = new ArrayList();
        int windowWidth, windowHeight;
        [Glade.Widget] Window window;
        [Glade.Widget] Statusbar statusbar;
        [Glade.Widget] Notebook notebook;
        [Glade.Widget] Expander expander;
        [Glade.Widget] EventBox expander_eventbox;
        [Glade.Widget] Viewport expander_vp;
        [Glade.Widget] HBox hbox_row1;
        [Glade.Widget] HBox hbox_row2;

        double PanBeginX = 0;
        double PanBeginHadj = 0;

        Format formatter;
        Format Formatter
        {
            get
            {
                if (formatter == null)
                    formatter = new Format(Settings.StatusFormat);
                return formatter;
            }
        }

        public static void Main(string[] args)
        {
            Gtk.Application.Init();
            new MpcSharpApp(args);
            Gtk.Application.Run();
        }


        public MpcSharpApp(string[] args)
        {
            gxml = Glade.XML.FromAssembly("gui.glade", "window", null);
            gxml.Autoconnect(this);

            // Create the connection object. This does NOT connect to mpd.
            conn = new Connection();

            // Add all small GuiWidgets
            SetGuiWidgets(hbox_row1, Settings.GuiRow1);
            SetGuiWidgets(hbox_row2, Settings.GuiRow2);

            // Restore old window size and position
            window.Move(Settings.WinPosX, Settings.WinPosY);
            window.Resize(Settings.WinWidth, Settings.WinHeight);

            // Read config and show all chosen tabs
            SetTabs();

            if (!Settings.Expanded)
                expander.Activate();

            // Show the window
            window.ShowAll();

            // Set small size request to prevent the status string to force the
            // main window to grow too large
            gxml.GetWidget("expander").SetSizeRequest(1, -1);

            // This is required to enable panning of the status string.
            // Otherwise the 'exapnder' steals all motion notify events.
            gxml.GetWidget("expander").AddEvents((int) Gdk.EventMask.ButtonMotionMask);

            // Register update delegates
            Connection.UpdateDelegate d = new Connection.UpdateDelegate(Update);
            conn.RegisterUpdateDelegate(d);

            // Force first update and refresh every 500 ms
            PollMpd();
            GLib.Timeout.Add(500, PollMpd);

            // GConf notifications
            Settings.StatusFormatChanged += StatusFormatNotifyEventHandler;
            Settings.GuiRow1Changed += GuiRow1NotifyEventHandler;
            Settings.GuiRow2Changed += GuiRow2NotifyEventHandler;
        }

        void GuiRow1NotifyEventHandler(object o, GConf.NotifyEventArgs args)
        {
            SetGuiWidgets(hbox_row1, Settings.GuiRow1);
        }

        void GuiRow2NotifyEventHandler(object o, GConf.NotifyEventArgs args)
        {
            SetGuiWidgets(hbox_row2, Settings.GuiRow2);
        }

        void Update()
        {
            // Status string and window title
            if (conn.prevStatusInfo.Music.SongID != conn.statusInfo.Music.SongID ||
                    conn.prevStatusInfo.PlaylistVersion != conn.statusInfo.PlaylistVersion)
            {
                UpdateStatusString();
                window.Title = conn.musicInfo.Filename;
            }

            // Statusbar
            string msg = "";
            switch(conn.statusInfo.State)
            {
                case StateStatus.Play:
                    MusicStatus ms = conn.statusInfo.Music;
                    msg = "Playing, " + ms.BitRate +  " kbps, " + ms.SampleRate + " Hz";
                    break;
                case StateStatus.Pause:
                    msg = "Paused";
                    break;
                case StateStatus.Stop:
                    msg = "Stopped";
                    break;
            }
            statusbar.Push(1, msg);

            // Updating database?
            if (conn.statusInfo.UpdatingDB)
            {
                statusbar.Push(1, "Updating database...");
            }
        }

        void StatusFormatNotifyEventHandler(object o, GConf.NotifyEventArgs args)
        {
            UpdateStatusString();
        }

        void SetTabs()
        {
            // Unregister all old UpdateDelegates
            IEnumerator e = visibleTabs.GetEnumerator();
            while (e.MoveNext())
            {
                if (e.Current is IUpdatable)
                    (e.Current as IUpdatable).UnregisterUpdateDelegate();
            }

            // Remove all old pages from the notebook
            int nPages = notebook.NPages;
            for (int i = 0; i< nPages; i++)
            {
                notebook.RemovePage(0);
            }

            visibleTabs.Clear();


            // Construct all requested tabs
            foreach (string tab in Settings.Tabs)
            {
                TabType type;
                try
                {
                type = (TabType) Enum.Parse(typeof(TabType), tab);
                }
                catch (ArgumentException)
                {
                    continue;
                }

                Widget widget = null;
                switch (type)
                {
                    case (TabType.Playlist):
                        widget = new Playlist(ref conn);
                        break;
                    case (TabType.Library):
                        widget = new Library(ref conn);
                        break;
                    case (TabType.Find):
                        widget   = new Find(ref conn);
                        break;
                    case (TabType.Cover):
                        widget = new Cover(ref conn);
                        break;
                    case (TabType.Playlists):
                        widget = new Playlists(ref conn);
                        break;
                }

                if (widget != null)
                {
                    visibleTabs.Add(widget);
                    notebook.AppendPage(widget, new Gtk.Label("_" + tab));
                }
            }
            notebook.ShowAll();
            notebook.Page = Settings.ActiveTab;

            // Set conn.client to null to force a proper first update
            conn.client = null;
        }

        void SetGuiWidgets(HBox hbox, string format)
        {
            // Remove all old widgets
            foreach (Widget w in hbox.Children)
            {
                if (w is IUpdatable)
                    (w as IUpdatable).UnregisterUpdateDelegate();
            }
            foreach (Widget w in hbox.Children)
            {
                hbox.Remove(w);
            }

            // Append the new ones
            MpcSharpWidget widget;
            foreach (string pair in format.Split(';'))
            {
                if (pair == "")
                    continue;
                bool visible = Convert.ToBoolean(pair.Split(':')[0]);
                if (visible)
                {
                    string name = pair.Split(':')[1];
                    widget = (MpcSharpWidget) Enum.Parse(typeof(MpcSharpWidget), name);
                    AddGuiWidget(hbox, widget);
                }
            }

            hbox.ShowAll();
        }

        void AddGuiWidget(Box box, MpcSharpWidget item)
        {
            Widget w;
            switch (item)
            {
                case MpcSharpWidget.Prev:
                    w = new PrevButton(ref conn);
                    box.PackStart(w, false, false, 0);
                    break;
                case MpcSharpWidget.Play:
                    w = new PlayButton(ref conn);
                    box.PackStart(w, false, false, 0);
                    break;
                case MpcSharpWidget.Stop:
                    w = new StopButton(ref conn);
                    box.PackStart(w, false, false, 0);
                    break;
                case MpcSharpWidget.Next:
                    w = new NextButton(ref conn);
                    box.PackStart(w, false, false, 0);
                    break;
                case MpcSharpWidget.Volume:
                    w = new VolumeScale(ref conn);
                    box.PackStart(w, true, true, 0);
                    break;
                case MpcSharpWidget.Random:
                    w = new RandomButton(ref conn);
                    box.PackStart(w, false, false, 0);
                    break;
                case MpcSharpWidget.Repeat:
                    w = new RepeatButton(ref conn);
                    box.PackStart(w, false, false, 0);
                    break;
                case MpcSharpWidget.Time:
                    w = new TimeLabel(ref conn);
                    box.PackStart(w, false, false, 0);
                    break;
                case MpcSharpWidget.Progress:
                    w = new ProgressScale(ref conn);
                    box.PackStart(w, true, true, 0);
                    break;
            }
        }

        bool PollMpd()
        {
            if (conn.client == null)
            {
                if (!conn.Connect())
                {
                    statusbar.Push(1, "Not connected");
                    Console.WriteLine("Connection failed");
                    return true;
                }

                if (!conn.Refresh())
                {
                    Console.WriteLine("Refresh failed");
                    return true;
                }

                // We are now connected. Make sure that the 'old' status values
                // are such that a first proper update is performed.
                conn.prevStatusInfo.Music.SongID = -1;
                conn.prevStatusInfo.Music.Song = -1;
                conn.prevStatusInfo.PlaylistVersion = 0;
                conn.prevStatsInfo.DBUpdateDate = DateTime.MinValue;
                statusbar.Push(1, "Connected");
            }
            else
            {
                if (!conn.Refresh())
                {
                    return true;
                }
            }

            // Run all registred update delegates
            conn.RunUpdateDelegates();

            return true;
        }

        void UpdateStatusString()
        {
            Formatter.MusicInfoFormat = Settings.StatusFormat;
            Label label = gxml.GetWidget("expander_label") as Label;
            string text = Formatter.MusicInfo(conn.musicInfo);
            text = text.Replace(System.Environment.NewLine, "");
            text = text.Replace("\\n", System.Environment.NewLine);
            text = text.Replace("\\t", "\t");
            label.Text = text;
            label.UseMarkup = true;
        }


        void ToggleTab(string title)
        {
            ArrayList tabs = new ArrayList();
            foreach (string tab in Settings.Tabs)
            {
                tabs.Add(tab);
            }
            if (tabs.IndexOf(title) > -1)
                tabs.Remove(title);
            else
                tabs.Add(title);
            if (tabs.Count == 0)
                // Can't remove the last tab
                return;
            Settings.Tabs = tabs.ToArray();

            SetTabs();
        }


        /*
         * Main gui callbacks
         */
        [GLib.ConnectBefore]
        void OnExpanderButtonRelease(object o, ButtonReleaseEventArgs args)
        {
            if (args.Event.Button == 3)
            {
                //Glade.XML gxml =
                //    Glade.XML.FromAssembly("gui.glade", "move_tab_menu", null);
                //gxml.Autoconnect(this);
                Menu tabConfigMenu = new Menu();
                MenuItem moveLeft = new MenuItem("Move left");
                MenuItem moveRight = new MenuItem("Move right");
                CheckMenuItem showPlaylist = new CheckMenuItem("Playlist");
                CheckMenuItem showLibrary = new CheckMenuItem("Library");
                CheckMenuItem showFind = new CheckMenuItem("Find");
                CheckMenuItem showCover = new CheckMenuItem("Cover");
                CheckMenuItem showPlaylists = new CheckMenuItem("Playlists");
                MenuItem showPreferences = new MenuItem("Preferences");

                tabConfigMenu.Append(moveLeft);
                tabConfigMenu.Append(moveRight);
                tabConfigMenu.Append(new SeparatorMenuItem());
                tabConfigMenu.Append(showPlaylist);
                tabConfigMenu.Append(showLibrary);
                tabConfigMenu.Append(showFind);
                tabConfigMenu.Append(showCover);
                tabConfigMenu.Append(showPlaylists);
                tabConfigMenu.Append(new SeparatorMenuItem());
                tabConfigMenu.Append(showPreferences);

                string[] tabs = Settings.Tabs as string[];
                showPlaylist.Active = System.Array.IndexOf(tabs, "Playlist") > -1;
                showLibrary.Active = System.Array.IndexOf(tabs, "Library") > -1;
                showFind.Active = System.Array.IndexOf(tabs, "Find") > -1;
                showCover.Active = System.Array.IndexOf(tabs, "Cover") > -1;
                showPlaylists.Active = System.Array.IndexOf(tabs, "Playlists") > -1;

                moveLeft.Activated += new EventHandler(OnMoveLeftActivate);
                moveRight.Activated += new EventHandler(OnMoveRightActivate);
                showPlaylist.Toggled += new EventHandler(OnShowPlaylistToggled);
                showLibrary.Toggled += new EventHandler(OnShowLibraryToggled);
                showFind.Toggled += new EventHandler(OnShowFindToggled);
                showCover.Toggled += new EventHandler(OnShowCoverToggled);
                showPlaylists.Toggled += new EventHandler(OnShowPlaylistsToggled);
                showPreferences.Activated += new EventHandler(OnShowOptions);

                tabConfigMenu.ShowAll();
                tabConfigMenu.Popup();
            }
            else if (args.Event.Button == 2)
            {
                Gtk.Grab.Remove(expander_eventbox);
            }
        }

        [GLib.ConnectBefore]
        void OnExpanderButtonPress(object o, ButtonPressEventArgs args)
        {
            if (args.Event.Button == 2)
            {
                PanBeginX = args.Event.XRoot;
                PanBeginHadj = expander_vp.Hadjustment.Value;
                Gtk.Grab.Add(expander_eventbox);
            }
        }

        [GLib.ConnectBefore]
        void OnExpanderActivated(object o, EventArgs args)
        {
            if (expander.Expanded)
            {
                int width, height;
                window.GetSize(out width, out height);
                Settings.WinWidth = width;
                Settings.WinHeight = height;
                expander.Child.HideAll();
            }
            else
            {
                expander.Child.ShowAll();
            }
        }

        void OnAfterExpanderActivated(object o, EventArgs args)
        {
            int width, height;
            window.GetSize(out width, out height);
            if (!expander.Expanded)
            {
                window.Resize(width, 1);
            }
            else
            {
                window.Resize(width, Settings.WinHeight);
            }
        }

        [GLib.ConnectBefore]
        void OnExpanderEventboxMotionNotify(object o, MotionNotifyEventArgs args)
        {
            double dx = args.Event.XRoot - PanBeginX;
            Adjustment hadj = expander_vp.Hadjustment;
            double new_hadj = System.Math.Min(PanBeginHadj - dx, hadj.Upper - hadj.PageSize);
            hadj.Value = new_hadj;
        }

        void OnDeleteEvent(object o, DeleteEventArgs args)
        {
            // Save window position
            int pos_x, pos_y;
            window.GetPosition(out pos_x, out pos_y);
            Settings.WinPosX = pos_x;
            Settings.WinPosY = pos_y;

            // Save window size
            if (expander.Expanded)
            {
                int width, height;
                window.GetSize(out width, out height);
                Settings.WinWidth = width;
                Settings.WinHeight = height;
            }

            // Save expanded state
            Settings.Expanded = expander.Expanded;

            // Save current tab
            Settings.ActiveTab = notebook.Page;

            Application.Quit();
            args.RetVal = true;
        }

        void OnKeyPress(object o, KeyPressEventArgs args)
        {
            Gdk.EventKey e = args.Event;

            // Quit - Escape
            if (e.Key == Gdk.Key.Escape)
            {
                Application.Quit();
            }
            // Settings - Ctrl-O
            else if (e.Key == Gdk.Key.o && (e.State & Gdk.ModifierType.ControlMask)
                    == Gdk.ModifierType.ControlMask)
            {
                new SettingsEditor();
            }
        }

        /*
         * Move tabs menu callbacks 
         */
        void OnMoveLeftActivate(object o, EventArgs args)
        {
            int newPos = System.Math.Max(0, notebook.Page -1);
            notebook.ReorderChild(notebook.CurrentPageWidget, newPos);
            SaveTabOrder();
        }

        void OnMoveRightActivate(object o, EventArgs args)
        {
            int newPos = System.Math.Min(notebook.Page + 1, notebook.NPages-1);
            notebook.ReorderChild(notebook.CurrentPageWidget, newPos);
            SaveTabOrder();
        }

        void SaveTabOrder()
        {
            ArrayList tabs = new ArrayList();
            for (int i=0; i < notebook.NPages; i++)
            {
                string label = notebook.GetTabLabelText(notebook.GetNthPage(i));
                if (label == "Playlist")
                    tabs.Add("Playlist");
                else if (label == "Find")
                    tabs.Add("Find");
                else if (label == "Cover")
                    tabs.Add("Cover");
                else if (label == "Library")
                    tabs.Add("Library");
            }
            Settings.Tabs = tabs.ToArray();
        }

        void OnShowPlaylistToggled(object o, EventArgs args)
        {
            ToggleTab("Playlist");
        }

        void OnShowLibraryToggled(object o, EventArgs args)
        {
            ToggleTab("Library");
        }

        void OnShowFindToggled(object o, EventArgs args)
        {
            ToggleTab("Find");
        }

        void OnShowCoverToggled(object o, EventArgs args)
        {
            ToggleTab("Cover");
        }

        void OnShowPlaylistsToggled(object o, EventArgs args)
        {
            ToggleTab("Playlists");
        }

        void OnShowOptions(object o, EventArgs args)
        {
            new SettingsEditor();
        }
    }
}
